/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi;

import java.util.HashSet;
import java.util.Set;

import org.osgi.framework.ServiceRegistration;

/**
 * Utility class that tracks a set of {@link ServiceRegistration ServiceRegistrations} and always for them to be safely
 * unregistered.
 * <p/>
 * <code>ServiceRegistrations</code> are tracked in a threadsafe manner, and importantly are unregistered without
 * holding any locks.
 * 
 * @see ServiceRegistration
 */
public final class ServiceRegistrationTracker {

    private final Object monitor = new Object();

    private Set<ServiceRegistration> registrations;

    /**
     * Tracks the supplied {@link ServiceRegistration}. This <code>ServiceRegistration</code> will be
     * {@link ServiceRegistration#unregister unregistered} during {@link #unregisterAll()}.
     * 
     * @param registration the <code>ServiceRegistration</code> to track.
     */
    public void track(ServiceRegistration registration) {
        synchronized (this.monitor) {
            if (this.registrations == null) {
                this.registrations = new HashSet<ServiceRegistration>();
            }
            this.registrations.add(registration);
        }
    }

    /**
     * Safely unregisters all the tracked <code>ServiceRegistrations</code>.
     */
    public void unregisterAll() {
        Set<ServiceRegistration> toUnregister = null;
        synchronized (this.monitor) {
            toUnregister = this.registrations;
            this.registrations = null;
        }
        if (toUnregister != null) {
            for (ServiceRegistration serviceRegistration : toUnregister) {
                serviceRegistration.unregister();
            }
        }
    }
}
