/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi;

import java.util.HashSet;
import java.util.Set;

import org.osgi.framework.BundleContext;
import org.osgi.util.tracker.ServiceTracker;

import com.springsource.util.common.SetProvider;



/**
 * <p>
 * This class provides a wrapper around the OSGi <code>ServiceTracker</code> class that 
 * lets you request all instance of a service at any time.
 * </p>
 *
 * <strong>Concurrent Semantics</strong><br />
 *
 * This class is thread safe.
 *
 * @author Christopher Frost
 * @param <T> parametric type
 * @since 1.5
 */
public class ServiceRegistryProvider<T> implements SetProvider<T> {

    private final ServiceTracker serviceTracker;

    private final Class<T> typeClazz;
    
    /**
     * Requires a bundle context to access the service registry trough. This should be the
     * bundle context belonging to the code using this service, not the bundle that is 
     * publishing the services of interest.
     * 
     * @param bundleContext of code using this service
     * @param clazz Class of service
     */
    public ServiceRegistryProvider(BundleContext bundleContext, Class<T> clazz) {
        this.typeClazz = clazz;
        this.serviceTracker = new ServiceTracker(bundleContext, this.typeClazz.getName(), null);
        this.serviceTracker.open();
    }

    /** 
     * {@inheritDoc}
     */
    public Set<T> getSet() {
        Object[] services = serviceTracker.getServices();
        Set<T> servicesSet = new HashSet<T>();
        if (services != null) {
            for (Object provider : services) {
                servicesSet.add(this.typeClazz.cast(provider));
            }
        }

        return servicesSet;
    }

    /**
     * Close the underlying <code>ServiceTracker</code> object.
     * 
     */
    public void close() {
        if (serviceTracker != null) {
            serviceTracker.close();
        }
    }
    
}
