/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest;

import java.util.List;

/**
 * Represents a bundle's <code>Bundle-ActivationPolicy</code> header.
 * <p />
 *
 * <strong>Concurrent Semantics</strong><br />
 * May not be thread-safe.
 */
public interface BundleActivationPolicy extends Parameterised {

    /**
     * An enumeration of the legal values for a bundle's activation policy, 
     * specified using the <code>Bundle-ActivationPolicy</code> header.
     * <p />
     */    
    enum Policy {
        /**
         * The bundle will be activated eagerly.
         */
        EAGER,
        /**
         * The bundle will be activated lazily.
         */
        LAZY;
    }

    /**
     * Returns the value of the <code>Bundle-ActivationPolicy</code> header. If no header
     * is specified returns the default value of {@link Policy#EAGER}.
     * @return the value of the <code>Bundle-ActivationPolicy</code> header.
     */
    Policy getActivationPolicy();
    
    /**
     * Sets the value of the <code>Bundle-ActivationPolicy</code> header.
     * 
     * @param policy the bundle's activation policy.
     */
    void setActivationPolicy(Policy policy);
    
    /**
     * Returns a list of the class names specified in the header's <code>exclude</code> directive. Returns an empty list
     * if the header has no <code>exclude</code> directive.
     * 
     * @return the list of exclusions
     */
    List<String> getExclude();
    
    /**
     * Returns a list of the class names specified in the header's <code>include</code> directive. Returns an empty list
     * if the header has no <code>include</code> directive.
     * 
     * @return the list of inclusions
     */
    List<String> getInclude();
}
