/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest;

import java.io.IOException;
import java.io.Reader;
import java.util.Dictionary;

import com.springsource.util.osgi.manifest.internal.StandardBundleManifest;
import com.springsource.util.osgi.manifest.parse.DummyParserLogger;
import com.springsource.util.osgi.manifest.parse.ParserLogger;
import com.springsource.util.parser.manifest.ManifestContents;

/**
 * This interface provides factory methods for creating empty bundle manifests and empty instances of the more complex
 * bundle headers.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Implementations of this interface must be thread safe.
 * 
 * @author Glyn Normington
 * @author Andy Wilkinson
 * @author Ben Hale
 * @since 1.0
 */
public class BundleManifestFactory {

    /**
     * Creates a new, empty {@link BundleManifest}.
     * 
     * @return the new <code>BundleManifest</code>.
     */
    public static BundleManifest createBundleManifest() {
        return createBundleManifest(new DummyParserLogger());
    }

    /**
     * Creates a new, empty {@link BundleManifest}.
     * 
     * @param parserLogger The parser logger to use when creating new headers in the manifest
     * @return the new <code>BundleManifest</code>.
     */
    public static BundleManifest createBundleManifest(ParserLogger parserLogger) {
        return new StandardBundleManifest(parserLogger);
    }

    /**
     * Creates a new {@link BundleManifest} derived from the supplied {@link Dictionary} of bundle manifest headers.
     * 
     * @param headers The <code>Dictionary</code> of headers
     * @return The <code>BundleManifest</code> derived from the <code>Dictionary</code>.
     */
    public static BundleManifest createBundleManifest(Dictionary<String, String> headers) {
        return createBundleManifest(headers, new DummyParserLogger());
    }

    /**
     * Creates a new {@link BundleManifest} derived from the supplied {@link Dictionary} of bundle manifest headers. The
     * supplied {@link ParserLogger} will be used to report problems encountered during parsing.
     * 
     * @param headers The <code>Dictionary</code> of headers
     * @param parserLogger The <code>ParserLogger</code> to be used to report parsing problems.
     * @return The <code>BundleManifest</code> derived from the <code>Dictionary</code>.
     */
    public static BundleManifest createBundleManifest(Dictionary<String, String> headers, ParserLogger parserLogger) {
        return new StandardBundleManifest(parserLogger, headers);
    }
    
    public static BundleManifest createBundleManifest(ManifestContents manifestContents, ParserLogger parserLogger) {
        return new StandardBundleManifest(parserLogger, manifestContents);
    }

    /**
     * Creates a new {@link BundleManifest}, reading its contents from the supplied {@link Reader} The supplied
     * {@link ParserLogger} will be used to report problems encountered during parsing.
     * 
     * @param reader The <code>Reader</code> of headers
     * @param parserLogger The <code>ParserLogger</code> to be used to report parsing problems.
     * @return The <code>BundleManifest</code> populated by reading the <code>Reader</code>.
     * @throws IOException if an error occurs reading the supplied <code>Reader</code>.
     */
    public static BundleManifest createBundleManifest(Reader reader, ParserLogger parserLogger) throws IOException {
        return new StandardBundleManifest(parserLogger, reader);
    }

    /**
     * Creates a new {@link BundleManifest}, reading its contents from the supplied {@link Reader}.
     * 
     * @param reader The <code>Reader</code> of headers
     * @return The <code>BundleManifest</code> populated by reading the <code>Reader</code>.
     * @throws IOException if an error occurs reading the supplied <code>Reader</code>.
     */
    public static BundleManifest createBundleManifest(Reader reader) throws IOException {
        return createBundleManifest(reader, new DummyParserLogger());
    }
}
