/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest.internal;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.springsource.util.osgi.manifest.Parameterised;
import com.springsource.util.osgi.manifest.Parseable;
import com.springsource.util.osgi.manifest.parse.HeaderDeclaration;
import com.springsource.util.osgi.manifest.parse.HeaderParser;

/**
 * <strong>Concurrent Semantics</strong><br />
 * Thread-safe.
 */
abstract class CompoundParseable<T extends Parameterised> implements Parseable {

    protected final HeaderParser parser;

    protected List<T> components;
    
    private static final String EMPTY_STRING = "";

    @SuppressWarnings("unchecked")
    CompoundParseable(HeaderParser parser) {
        this.parser = parser;
        this.components = new ArrayList();
    }

    abstract List<HeaderDeclaration> parse(String parseString);

    abstract T newEntry(String name);

    /**
     * {@inheritDoc}
     */
    public void resetFromParseString(String parseString) {

        this.components.clear();

        if (parseString != null) {
            List<HeaderDeclaration> headers = parse(parseString);

            for (HeaderDeclaration header : headers) {
                for (String name : header.getNames()) {
                    T component = newEntry(name);
                    component.getAttributes().putAll(header.getAttributes());
                    component.getDirectives().putAll(header.getDirectives());
                    this.components.add(component);
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String toParseString() {
        if (this.components.isEmpty()) {
            return null;
        }
        StringBuilder builder = new StringBuilder();

        Iterator<T> components = this.components.iterator();
        while (components.hasNext()) {
            builder.append(components.next().toParseString());
            if (components.hasNext()) {
                builder.append(",");
            }
        }
        return builder.toString();
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        String parseString = toParseString();
        if (parseString == null) {
            return EMPTY_STRING;
        } else {
            return parseString;
        }
    }

    protected T add(String name) {
        T newComponent = newEntry(name);
        this.components.add(newComponent);
        return newComponent;
    }
}
