/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest.parse;

import java.util.List;

/**
 * Strategy for parsing OSGi manifest headers.<p/>
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * Implementations need not be threadsafe.
 * 
 * @author Rob Harrop
 * @author Glyn Normington
 * @author Andy Clement
 * @since 1.0
 */
public interface HeaderParser {

    /**
     * Parses the supplied import/export package header text and returns the list of corresponding
     * {@link HeaderDeclaration}s.
     * 
     * @param header the header text to parse.
     * @param headerType 
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parsePackageHeader(String header, String headerType);

    /**
     * Parses the supplied dynamic import header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}. Dynamic imports allow wildcarded package names.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parseDynamicImportPackageHeader(String header);

    /**
     * Parses the supplied require bundle header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parseRequireBundleHeader(String header);

    /**
     * Parses the supplied fragment host header text and returns the corresponding
     * {@link HeaderDeclaration HeaderDeclaration}.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}.
     * 
     */
    HeaderDeclaration parseFragmentHostHeader(String header);

    /**
     * Parses the supplied bundle symbolic name header text and returns the corresponding
     * {@link HeaderDeclaration HeaderDeclaration}.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration} for the Bundle-SymbolicName.
     */
    HeaderDeclaration parseBundleSymbolicName(String header);

    /**
     * Parses the supplied import bundle header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parseImportBundleHeader(String header);

    /**
     * Parses the supplied Import-Library header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parseImportLibraryHeader(String header);

    /**
     * Parses the supplied Library-SymbolicName header text and returns the list of corresponding
     * {@link HeaderDeclaration HeaderDeclarations}.
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}.
     */
    HeaderDeclaration parseLibrarySymbolicName(String header);
    
    /**
     * Parses the supplied <code>Web-FilterMappings</code> header text and returns the list
     * of corresponding {@link HeaderDeclaration HeaderDeclarations}.
     * @param header 
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parseWebFilterMappingsHeader(String header);
    
    /**
     * Parses the supplied <code>Bundle-ActivationPolicy</code> header text and returns the
     * corresponding {@link HeaderDeclaration}.
     * @param header 
     * @return the {@link HeaderDeclaration}.
     */
    HeaderDeclaration parseBundleActivationPolicy(String header);
    
    
    /**
     * Parses the supplied header text and returns the list of corresponding {@link HeaderDeclaration}.
     * Format for the header is expected to follow OSGi 3.2.4 "Common Header Syntax"
     * 
     * @param header the header text to parse.
     * @return the {@link HeaderDeclaration}s.
     */
    List<HeaderDeclaration> parseHeader(String header);
}
