/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest.parse.standard;

/**
 * Encapsulates a single problem found during header parsing. The problem
 * includes a kind {@link HeaderProblemKind}, together with the start and end
 * offset for the problem in the data and some inserts that would be included in
 * the formatted form of this message. The toString() methods like to have the
 * context passed in so that they can produce nice results - the source context
 * encapsulates the source data that was being processed together with useful
 * information from that (eg. where the line breaks are - but that is less
 * useful in this situation where all problems are on the same line).
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Andy Clement
 */
public class HeaderProblem {

	private HeaderProblemKind problemKind;
	private int startoffset;
	private int endoffset;
	private String[] inserts;

	public HeaderProblem(HeaderProblemKind problem, int startoffset, int endoffset, String... inserts) {
		this.problemKind = problem;
		this.startoffset = startoffset;
		this.endoffset = endoffset;
		this.inserts = inserts;
	}

	public String toString(SourceContext context) {
		return problemKind.format(context.getLine(startoffset), context.getColumn(startoffset), inserts);
	}

	public String toStringWithContext(SourceContext context) {
		int relevantLine = context.getLine(startoffset);
		String relevantLineData = context.getLineAsString(relevantLine);
		String theMessage = toString(context);
		StringBuilder beautifulMessage = new StringBuilder();
		beautifulMessage.append(relevantLineData).append('\n');
		int scol = context.getColumn(startoffset);
		int ecol = context.getColumn(endoffset);
		int where = 0;
		for (where = 0; where < scol; where++) {
			beautifulMessage.append(' ');
		}
		beautifulMessage.append('^');
		if (ecol != scol) {
			where++;
			for (; where < ecol; where++) {
				beautifulMessage.append(' ');
			}
			beautifulMessage.append('^');
		}
		beautifulMessage.append('\n');
		beautifulMessage.append(theMessage);
		return beautifulMessage.toString();
	}

	@Override
	public String toString() {
		return this.problemKind.format(0, 0, inserts);
	}

	public HeaderProblemKind getKind() {
		return this.problemKind;
	}

	public String[] getInserts() {
		String[] insertCopy = new String[inserts.length];
		System.arraycopy(inserts, 0, insertCopy, 0, inserts.length);
		return insertCopy;
	}

	public int getStartOffset() {
		return this.startoffset;
	}

	public int getEndOffset() {
		return this.endoffset;
	}

	public boolean isSeverity(Severity severity) {
		return this.problemKind.isSeverity(severity);
	}

}
