/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest.parse.standard;

import java.util.List;

import com.springsource.util.osgi.manifest.parse.HeaderDeclaration;

/**
 * HeaderDeclarations are constructed through a visitor pattern. As the parser
 * processes data it calls methods on the visitor interface. The standard
 * implementation is {@link StandardHeaderVisitor} but other visitors can be
 * plugged in. For example, the {@link DebugVisitor} produces diagnostics about
 * the visiting process. A validation visitor could be plugged in to verify
 * package names or attribute names and values.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Andy Clement
 */
public interface HeaderVisitor {

	/**
	 * Visit a new directive (parsed from input of the form name:=value)
	 * @param name 
	 * @param value 
	 */
	void visitDirective(String name, String value);

	/**
	 * Visit a new attribute (parsed from input of the form name=value)
	 * @param name 
	 * @param value 
	 */
	void visitAttribute(String name, String value);

	/**
	 * @return a list of all HeaderDeclarations if this visitor has been collecting the information. Empty list if nothing has been
	 *         collected
	 */
	List<HeaderDeclaration> getHeaderDeclarations();

	/**
	 * @return the first (and probably only) HeaderDeclaration, otherwise null
	 */
	HeaderDeclaration getFirstHeaderDeclaration();

	/**
	 * Called when a clause ends. Clauses are comma separated: header=clause (',' clause)*. It is a sign that a new clause is about
	 * to start.
	 */
	void clauseEnded();

	/**
	 * Called to visit a symbolic name. These are of the form: "symbolic-name :: = token('.'token)*". So basically a sequence of dot
	 * separated tokens.
	 * @param name 
	 */
	void visitSymbolicName(String name);

	/**
	 * Called to visit a unique name. These are of the form: "unique-name :: = identifier('.'identifier)*". So basically a sequence
	 * of dot separated identifiers..
	 * @param name 
	 */
	void visitUniqueName(String name);

	/**
	 * Called to visit a wildcard name. These are similar to a unique name but wildcards are permitted. These are valid wildcarded
	 * names: '*' or 'com.foo.*' or 'com.foo.goo'.
	 * @param name 
	 */
	void visitWildcardName(String name);

	/**
	 * Called when the end of the input has been reached.
	 */
	void endvisit();

	/**
	 * Called ahead of parsing a new header
	 */
	void initialize();

}
