/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.osgi.manifest.parse.standard;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.springsource.util.osgi.manifest.parse.HeaderDeclaration;

/**
 * Standard implementation of the HeaderVisitor. This implementations collects
 * up the names, attributes and directives visited and for each clause creates a
 * HeaderDeclaration. These are then available at the end of the parse through
 * getHeaderDeclaration() and getFirstHeaderDeclaration()
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * Threadsafe.
 * 
 * @author Andy Clement
 */
public class StandardHeaderVisitor implements HeaderVisitor {

	// The complete list of HeaderDeclarations discovered during the parse
	private List<HeaderDeclaration> headerDeclarations;

	// Collections populated as names/directives/attributes are visited
	private List<String> names;
	private Map<String, String> directives;
	private Map<String, String> attributes;

	public void visitAttribute(String name, String value) {
		attributes.put(name, value);
	}

	public void visitDirective(String name, String value) {
		directives.put(name, value);
	}

	public void visitSymbolicName(String name) {
		names.add(name);
	}

	public void visitUniqueName(String name) {
		names.add(name);
	}

	public void visitWildcardName(String name) {
		names.add(name);
	}

	public void endvisit() {
	}

	/**
	 * Record a new HeaderDeclaration and reset the collections accumulating names, attributes and directives.
	 */
	public void clauseEnded() {
		headerDeclarations.add(new StandardHeaderDeclaration(names, attributes, directives));
		
		names.clear();
		directives.clear();
		attributes.clear();
	}

	/**
	 * @return the first header declaration or null if there are none.
	 */
	public HeaderDeclaration getFirstHeaderDeclaration() {
		if (headerDeclarations.size() == 0) {
			return null;
		} else {
			return headerDeclarations.get(0);
		}
	}

	public List<HeaderDeclaration> getHeaderDeclarations() {
		return headerDeclarations;
	}
	
	/**
	 * Called ahead of parsing a header - allowing state to be reset in a reused visitor
	 */
	public void initialize() {
		headerDeclarations = new ArrayList<HeaderDeclaration>();
		names = new ArrayList<String>();
		directives = new HashMap<String, String>();
		attributes = new HashMap<String, String>();
	}
}
